
#include "MAX7219.h"

//the opcodes for the MAX7219
#define OP_NOOP   0
#define OP_DIGIT0 1
#define OP_DIGIT1 2
#define OP_DIGIT2 3
#define OP_DIGIT3 4
#define OP_DIGIT4 5
#define OP_DIGIT5 6
#define OP_DIGIT6 7
#define OP_DIGIT7 8
#define OP_DECODEMODE  9
#define OP_INTENSITY   10
#define OP_SCANLIMIT   11
#define OP_SHUTDOWN    12
#define OP_DISPLAYTEST 15

//---------------------------------------------------------------------------------------
MAX7219::MAX7219(int dataPin, int clkPin, int csPin) {
//---------------------------------------------------------------------------------------
    SPI_MOSI=dataPin;
    SPI_CLK=clkPin;
    SPI_CS=csPin;
    pinMode(SPI_MOSI,OUTPUT);
    pinMode(SPI_CLK,OUTPUT);
    pinMode(SPI_CS,OUTPUT);
    digitalWrite(SPI_CS,HIGH);
    SPI_MOSI=dataPin;
	spiTransfer(OP_DISPLAYTEST, 0);
    //set Scan Limit to 7 (8 digits)
    spiTransfer(OP_SCANLIMIT, 7);
	//decode is done in source
	spiTransfer(OP_DECODEMODE, 0);
	clearDisplay();
	//we go into shutdown-mode on startup
	shutdown(true);
}

//---------------------------------------------------------------------------------------
void MAX7219::shutdown(bool b) {
//---------------------------------------------------------------------------------------
    if(b)
        spiTransfer(OP_SHUTDOWN, 0);
    else
        spiTransfer(OP_SHUTDOWN, 1);
}

//---------------------------------------------------------------------------------------
void MAX7219::setIntensity(int intensity) { //0..15
//---------------------------------------------------------------------------------------
    spiTransfer(OP_INTENSITY, intensity & 0x0f);
}

//---------------------------------------------------------------------------------------
void MAX7219::clearDisplay(void) {
//---------------------------------------------------------------------------------------
    for(int i=1; i<=8; i++) {
        spiTransfer(i, 0);
    }
}

//---------------------------------------------------------------------------------------
void MAX7219::setDigit(int digit, byte value, boolean dp) {
//---------------------------------------------------------------------------------------
    byte v;
    if(digit<0 || digit>7 || value>15)
        return;
    v = pgm_read_byte_near(charTable + value); 
    if(dp){
		v |= B10000000;
	}
    spiTransfer(digit+1, v);
}

//---------------------------------------------------------------------------------------
void MAX7219::setChar(int digit, char value, boolean dp) {
//---------------------------------------------------------------------------------------
    byte index, v;

    if(digit<0 || digit>7)
        return;
    index=(byte)value;
    if(index >127) {
        //no defined beyond index 127, so we use the space char
        index=32;
    }
    v=pgm_read_byte_near(charTable + index); 
    if(dp){
        v |= B10000000;
	}
    spiTransfer(digit+1, v);
}
//---------------------------------------------------------------------------------------
void MAX7219::spiTransfer(volatile byte opcode, volatile byte data) {
//---------------------------------------------------------------------------------------
    //enable the line 
    digitalWrite(SPI_CS,LOW);
    //shift out opcode and data 
    shiftOut(SPI_MOSI,SPI_CLK,MSBFIRST,opcode);
    shiftOut(SPI_MOSI,SPI_CLK,MSBFIRST,data);
    //latch the data onto the display
    digitalWrite(SPI_CS,HIGH);
}
