/*
Linearanschlag - Zabex 2016
   
Pin assignment:
==============
PD0 Rx Keyboard
PD1 Tx Display
PB1 9   Proximity Switch  OC1B   
PB2 10  Stepper Enable 
PB3 11  Stepper Dir       MOSI
PB4 12  Stepper Step      MISO

 
/*
 *  0..9 Zifferneingabe. Dezimalpunkt fest auf 2. Stelle
 *  * or # Start
 *  A Alles löschen (neu Anfangen)
 *  B neue Referenzfahrt
 *  C Calibrate (Offset Eingabe) 
 *  D Delete   (Backspace)
 */

/*
 
Timerwerte für Geschwindigkeiten in Tabelle
16 microsteps, 200Steps/Umdrehung  3200µSteps/Umdrehung

HTD-M3, 20Teeth Zahnriemenscheibe = 60mm je 3200 µSteps
 
Beschleunigung auf Vmax in ca. 100mm =>  5333 µSteps
Mindestgeschwindigkeit V0 auf 32Hz  

CPU=16MHz  T1 Vorteiler /8 => 2MHz   OCR1A=65535=>30,5Hz ; OCR1A=188=>10456Hz



Referenzfahrt:
  Referenztaster ist 28,5cm vom Ende weg und wird bis zum Ende gedrückt.
  Realisiert mit induktivem Näherungsschalter der das vorbeifahrende Alu-Profil erkennt.
  Referenzfahrt immer schleichend auf schließenden Taster.

*/
#include <EEPROM.h>

//Some bits
#define BIT0 1
#define BIT1 2
#define BIT2 4
#define BIT3 8
#define BIT4 16
#define BIT5 32
#define BIT6 64
#define BIT7 128

//some data types
#define U8 byte
#define U16 unsigned int 
#define I16 int 
#define U32 unsigned long 
#define I32 signed long 

//Signal definitions
#define ENABLE_L      PORTB &= ~BIT2   //Pin 10
#define ENABLE_H      PORTB |=  BIT2
#define DIR_NEG       PORTB &= ~BIT3   //Pin 11
#define DIR_POS       PORTB |=  BIT3
#define STEP_L        PORTB &= ~BIT4   //Pin 12
#define STEP_H        PORTB |=  BIT4
#define REF_SENSOR_ON_METALL  (!(PINB & BIT1))    //PIN 9 

//Limits
#define posMin 50     //5mm
#define posMax 6180   //61.8cm
#define VmaxPhys 10600  //MicroSteps/s


// volatile global variables
volatile U16 v=32;
volatile bool standstill;
volatile U8  state=0;
volatile U16 StepsToTravel=0;
volatile U16 stepCount;
volatile bool triggerOnKeyPressed=false;
volatile bool triggerOnKeyReleased=false;

// global variables
U16 tmr=0;
U16 targetPos_mm10; //in 1/10 mm.
U16 targetPos=1234; //in increments. Possible StepsToTravel with 60mm/turn (HTD-M3, 20Teeth), 3200 Steps/turn (16Microstates)
                    //results in Step 0..32960 for 0..618mm
U16 currentPosition; //current position in increments                   
U16 Vmax;
U16 refOffset = 16000; //30cm
U16 refPos = 16000;    //RefOffset at last reference movement
I16 diff; 
bool calibrate=false;
bool startReferencing=false; 
bool moveToTargetPos = false;   
bool redrawDisplay = false;  // set true if display needs update due to keyPress
bool autoClear = false;
bool keyPressed = false;

//--------------------------------------------------------------------------------------------------------
void setup() {
//--------------------------------------------------------------------------------------------------------
    pinMode(9, INPUT_PULLUP);
    pinMode(10, OUTPUT);
    pinMode(11, OUTPUT);
    pinMode(12, OUTPUT);
    pinMode(13, OUTPUT);
    // initialize serial:
    Serial.begin(9600);

    TCCR1A = 0;
    TCCR1B = (1<<WGM12)|(1<<CS11); //CTC-Mode  /8 = 2MHz Takt
    TCCR1C = 0;
    OCR1A=60000;
    TIMSK1=(1<<OCIE1A);
    //TIMSK0=0;
    TIMSK2=0;

    refOffset = 256 * EEPROM.read(1) + EEPROM.read(0);
    if ((refOffset > 25000) || (refOffset < 5000)) refOffset = 15000; //Default value
    delay(600); //wait until terminal is ready for sure
    Serial.println("*PRESS B"); //B for referencing
    Serial.flush();
}

//--------------------------------------------------------------------------------------------------------
void moveSteps(U16 n){
//--------------------------------------------------------------------------------------------------------
    noInterrupts();
      StepsToTravel = n;
      standstill=false;
      stepCount = 0;
    interrupts();
}

//--------------------------------------------------------------------------------------------------------
void loop() {
//--------------------------------------------------------------------------------------------------------
U16 lcdVal;
U8 digit;
U16 mask;
bool leadingZero;

  if (redrawDisplay && (state >= 20)){ //update LCD (if not in reference move)
    redrawDisplay=0;    
    Serial.write("+"); //Mitte Display (rechten 4 Digits) 
 
    lcdVal = targetPos_mm10;
    mask=1000;
    leadingZero=true;
    for (byte i=0; i<4; i++){    
        digit = lcdVal/mask;        
        if (2==i)   Serial.write('.');
        if (digit > 0) leadingZero=false;
        if (leadingZero){
             Serial.write(' ');
        }else    {
            Serial.write(digit+48);
        }
        lcdVal%=mask;        
        mask /=10;
    }
  }
  digitalWrite(13, REF_SENSOR_ON_METALL);  //LED shows proximity switch state
  
  switch(state){    //1..17: reference movement
    case 0: 
        ENABLE_H; //motor power off
        if (startReferencing) 
            state++;
        break;
    case 1: //	inductive proximity switch sees metal? Defines move direction.
        ENABLE_L; //Turn motor ON
         Vmax = VmaxPhys;
        Serial.println("*RE.F. . ");
        if (REF_SENSOR_ON_METALL){
            state=10; //move POS
        }else{
            state=2;  //move NEG
        }
        v=25;
        break;
        
    case 2: //Refkey is not pressed. Move NEG direction (towards ref key)
        DIR_NEG;
        triggerOnKeyPressed  = true; //Brake when Metall reached; Set stepCount=0 at Triggerpos.
        moveSteps(50000);
        state++;
        break;
    case 3: //wait until standstill 
      if (standstill){
        delay(200); //Debounce key, calm dowm mechanic
        //Now I am in a "on metall" position. 
        //Change direction and move to a position where the proximity switch is released
        state++; 
      }
      break;
    case 4: //Move POS direction to switch-release-position 
      DIR_POS;
      moveSteps(stepCount + 160); //3mm über Endschalter hinaus (200[Steps/turn] * 16[uSteps/step] / 60[mm/turn] * 3[mm] )
      state=15;
      break;
      
    //----------------  
      
    case 10: //Move POS direction until ref switch is released
      DIR_POS;
      triggerOnKeyReleased  = true; //Brake when proximity swicht leaves "on metall" region
      moveSteps(50000);
      state++;
      break;
    case 11: //wait until standstill 
      if (standstill){
        delay(200); //calm dowm mechanic
        state++; //move other direction
      }
      break;
    case 12: //move NEG close to "on metall" position
      DIR_NEG;
      if (stepCount > 160){
        moveSteps(stepCount - 160);//3mm before "on metall" (3200/60*3)
      }
      state=15;
      break;
      
     //----------- 
     
    case 15:  //await standstill
      if (standstill){
        delay(200); //calm down mechanic
        state++; 
      }
      break;
    case 16:  //move slowly NEG to an "on metall" position 
      DIR_NEG;
      triggerOnKeyPressed = true; //Brake when proximity switch sees metall
      Vmax=1;
      moveSteps(3000); 
      state++;
      break;
    case 17: //await standstill
      if (standstill){
        currentPosition = refOffset - stepCount;
        Serial.println("*P0S ");
        Vmax = VmaxPhys;
        refPos = refOffset;
        targetPos_mm10 = currentPosition *6L / 32;
        redrawDisplay = true;
        autoClear = true;
        startReferencing=false;
        state=20; //Fertig
      }
      break;


    case 20: //ready for commands. Motor is powered on and not moving.
        if (startReferencing){
            state=1;
        }
        if (calibrate){
            calibrate=false;
            state=50;
        }
        if (moveToTargetPos){
            moveToTargetPos=false;
            state=100;
        }
        break;

    case 50: //calibrate Pos
        Serial.println("*PRESS C");
        delay(200); //wait until CR behind C has arrived and debounce key.
        state ++;
        break;
    case 51:
        keyPressed = false;
        calibrate = false;
        state ++;
        break;
    case 52: //wait for confirmation key (C) or any other key to cancel
        if (keyPressed){
          state ++; 
        }
        break;
           
    case 53: //calibrate Pos
        if (currentPosition > 5333){ //>100mm
          Serial.println("*ERROR");
          delay(1500);
          redrawDisplay = true;
        }else   
        if (calibrate) { // C pressed second time
          diff = currentPosition;
          currentPosition = (I32)targetPos_mm10  *32L/6/10;  //  3200/60/10/10  Input in 1/100mm (instead of 1/10)
          diff -= currentPosition; 
          
          refOffset = refPos - diff;
          EEPROM.update(0,  refOffset & 255);
          EEPROM.update(1,  (refOffset / 256) & 255);

          targetPos_mm10 = (I32)currentPosition * 6L /32;
          Serial.println("*CAL don.e");
          delay(1500);
          redrawDisplay = true;
        }else{
          Serial.println("*CANCELED");
          delay(1500);
          redrawDisplay = true;
        }
        calibrate = false;
        state=20;
        break;
        
    case 100: //Move to target position  
        //Movement always ends (like reference positioning) with NEG movement to avoid reversal backlash
        Serial.println("*P0S ");
        redrawDisplay = true;
        targetPos = (I32)targetPos_mm10  *32L/6;  //  3200/60/10
        Vmax = VmaxPhys;
        if (targetPos > currentPosition){
            DIR_POS;
            moveSteps(targetPos - currentPosition + 32); //move  0.6mm too far
            state=101;
        }else{
            DIR_NEG;
            moveSteps(currentPosition - targetPos);
            state=103;
        }
        currentPosition = targetPos;
        break;
    case 101: //wait until in "too far" position 
        if (standstill) state++;
        break;
    case 102: //move back 32 microsteps:    
        DIR_NEG;
        moveSteps(32);
        state++;
        break;
    case 103: //wait until in position:    
        if (standstill) state++;
        break;
    case 104:
        autoClear = true;
        state=20;
        break;
  }//switch
}



/*
  SerialEvent occurs whenever a new data comes in the
 hardware serial RX.  This routine is run between each
 time loop() runs, so using delay inside loop can delay
 response.  Multiple bytes of data may be available.
 */
//--------------------------------------------------------------------------------------------------------
void serialEvent() {
//--------------------------------------------------------------------------------------------------------
int len;  
    while (Serial.available()) {
        // get the new byte:
        char inChar = (char)Serial.read();
        if (inChar > 31) keyPressed = true; //ignore linefeeds
        if ('A' == inChar) { //clear All 
            targetPos_mm10 = 0;
            redrawDisplay=true;
            autoClear = false; 
        }else
        if ('B' == inChar) { // start referencing
            targetPos_mm10 = 0;
            startReferencing=true; 
            redrawDisplay=false;
            Serial.println("*REF");
            Serial.flush();
        }else
        if ('C' == inChar){           //calibrate
            calibrate=true;
            redrawDisplay=true;
        }else
        if ('D' == inChar){             // DEL (backspace)  
            targetPos_mm10 /= 10;
            autoClear = false; 
            redrawDisplay=true;
        }else
            
        if ( (inChar>='0') && (inChar<='9')) {
            if (autoClear){
              targetPos_mm10 = 0;
              autoClear = false; 
            }
            if (targetPos_mm10 < 1000){
              targetPos_mm10 = targetPos_mm10 * 10 + (inChar - '0') ;
            }
            redrawDisplay=true;
        }  
        // if the incoming character is a "#", set a flag
        // so the main loop can do something about it:
        if (('#' == inChar) || ('*' == inChar)) {
            moveToTargetPos = true;
            if (targetPos_mm10 > posMax){
                targetPos_mm10 = posMax;
                redrawDisplay=true;
                moveToTargetPos = false;
            }
            if (targetPos_mm10 < posMin){
                targetPos_mm10 = posMin;
                redrawDisplay=true;
                moveToTargetPos = false;
            }
        }
    }//while
}

const U16 tabx[]  ={0,
    10,33,34,35,36,37,39,41,43,45,47,49,51,54,57,60,64,68,72,77,
    82,88,95,103,112,122,134,148,164,183,206,234,269,313,369,442,541,678,876,1178,
    1671,2559,4413,5666,6784,10456,
    60000};

const U16 taby[]  ={//62499,  42552,
    15000,6000,
    3622,3532,3447,3366,3293,3158,3038,2931,2831,2746,2665,2593,2492,2405,2324,2228,2144,2069,1987,
    1912,1833,1751,1671,1592,1517,1439,1362,1287,1213,1138,1063,988,912,837,762,687,612,537,462,
    387,312,237,210,198,188,
    188};

#define TABLELEN 47

//--------------------------------------------------------------------------------------
ISR(TIMER1_COMPA_vect){//timer1 interrupt
//--------------------------------------------------------------------------------------
static U16 dx=10;
static U16 dy=10000;
static U16 xh = tabx[2];
static U16 xl = tabx[1];
static U16 yh = taby[2];
static U16 yl = taby[1];
static U8  index=2;

  if (StepsToTravel > 0){
    STEP_H;
    StepsToTravel --;
 
    if (triggerOnKeyPressed){
      if (REF_SENSOR_ON_METALL){
        triggerOnKeyPressed = false;
        stepCount=0;
        StepsToTravel = v; //deceleration ramp
      }
    }else
    if (triggerOnKeyReleased){
      if (!REF_SENSOR_ON_METALL){
        triggerOnKeyReleased = false;
        stepCount=0;
        StepsToTravel = v; 
      }
    }
    
    v++;
    if (v > StepsToTravel)  v = StepsToTravel;
    if (v > Vmax) v=Vmax;
    
    
    //linear interpolate velocity
    if ((v >= xh) && (index < TABLELEN)){
        xl=xh;
        yl=yh;
        index++;
        xh = tabx[index];
        yh = taby[index];
        dx=xh - xl;
        dy=yl - yh;
    }else
    if ((v < xl) && (index >0)){
        xh=xl;   
        yh=yl;
        index--;
        xl = tabx[index];
        yl = taby[index];
        dx=xh - xl;
        dy=yl - yh;
    }

    OCR1A =(U16)(yl - (I32)(v - xl) * dy / dx );
        
    stepCount++;

    STEP_L;
  }else standstill=1;
}


